 package es.rediris.papi;



import java.util.HashMap;
import java.util.Map;

import javax.crypto.Cipher;
import javax.crypto.spec.SecretKeySpec;

import org.bouncycastle.util.encoders.Base64;

/**
 * AESTool is a helper class for encrypting/decrypting messages using AES algorithm in ECB
 * mode. Also, it uses PKCS5 padding when makes its operations with messages.<br/>
 * Keys are 128-bit, 192-bit or 256-bit blocks, and you have to associate a name for their.
 * So, when you encrypt/decrypt you have to specify which key (through its name) you want to 
 * use.<br/>
 */
public class AESTool {
	private Map<String,byte[]> keys;

	/**
	 * Generate an <tt>AESTool</tt> object
	 *
	 */
	public AESTool() {
        keys=new HashMap<String,byte[]>();
	}
	
	/**
	 * Add a valid key setting its name.
	 * @param name The name of the key
	 * @param key The array of bytes defining the key
	 * @return <tt>true</tt> if the key is valid and it wasn't previously added, <tt>false</tt>
	 * in other case
	 */
	public boolean addKey(String name,byte[] key) {
		if (key==null||name==null||keys.containsKey(name))
			return false;
		//AJAY: Using && instead of ||
		if (key.length!=16&&key.length!=24&&key.length!=32) {
			return false;
		}
		keys.put(name,key);
		return true;
	}
	
	/**
	 * Delete a previously added key
	 * @param name The name of the key
	 * @return <tt>true</tt> if it has removed, <tt>false</tt> in other case
	 */
	public boolean delKey(String name) {
		if (name==null||!keys.containsKey(name))
			return false;
		keys.remove(name);
		return true;
	}
	
	/**
	 * Delete all added keys
	 *
	 */
	public void clearKeys() {
		keys.clear();
	}
	
	/**
	 * Get an array with the name of all added keys
	 * @return An array with their names
	 */
	public String[] getNameKeys() {
		String[] res=new String[keys.size()];
		res=keys.keySet().toArray(res);
		return res;
	}
	
	/**
	 * Encrypt a message using the specified key.
	 * @param nameKey The name of the key you want to use
	 * @param data The message you want to encrypt
	 * @return The ciphered message in base 64, or <tt>null</tt> if it has happened any error
	 * encrypting
	 */
    public String encode(String nameKey,String data) {
		try {
			Cipher cipher = Cipher.getInstance("AES/ECB/PKCS5Padding");
			//setup key
			byte[] keyBytes= (byte[])keys.get(nameKey);
			SecretKeySpec keySpec = new SecretKeySpec(keyBytes, "AES");
			cipher.init(Cipher.ENCRYPT_MODE,keySpec);
			byte[] rawData = data.getBytes();
			byte [] results = cipher.doFinal(rawData);
			//return new String(base64Encoder.encode(results));
			String output = new String(Base64.encode(results));
			return output;
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
    }
    
    /**
     * Decrypt a message using the specified key
     * @param nameKey The name of the key you want to use
     * @param data The ciphered message in base 64
     * @return The "plain" message, or <tt>null</tt> if it has happened any error
     */
    public String decode(String nameKey,String data) {
		try {
			Cipher cipher = Cipher.getInstance("AES/ECB/PKCS5Padding");
			//setup key
			byte[] keyBytes= (byte[])keys.get(nameKey);
			SecretKeySpec keySpec = new SecretKeySpec(keyBytes, "AES");
			cipher.init(Cipher.DECRYPT_MODE,keySpec);
			//byte[] rawData = base64Decoder.decodeBuffer(data);
			byte[] rawData = Base64.decode(data.getBytes());
			byte [] results = cipher.doFinal(rawData);
			return new String(results);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
    }
    
    public boolean hasKey (String nameKey) {
    	return this.keys.containsKey(nameKey);
    }
    
}
