package es.rediris.papi;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.Vector;
import org.apache.log4j.Logger;
import org.apache.log4j.xml.DOMConfigurator;

import javax.security.auth.login.LoginException;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;


public class PAPIPoA {

    /** Initial status */
    public final static int INIT = 0;
    /** Login query has been accepted */
    public final static int LOGIN_ACCEPTED = 1;
    /** Access query has been accepted */
    public final static int ACCESS_ACCEPTED = 2;
    /** Access denied due to invalid credentials */
    public final static int ACCESS_UNAUTHORIZED = 3;
    /** Access denied due to relevant errors in data */
    public final static int ACCESS_FATAL = 4;
    /** Login query rejected */
    public final static int LOGIN_REJECTED = 5;
    /** Logout query accepted */
    public final static int LOGOUT_ACCEPTED = 6;

    public static Logger log;
    
    private String configFileName = "";
    private static Properties papiConf = null;
    private Ciphers ciphers = null;
    private Map<String,String[]> parameters;
    private static String serviceID;
    private static String location;
    private static Map<String,Vector<String>> cookieRegistry;
    private BufferedWriter cookieFileWriter;
    private static HashMap<String,HashMap<String,String>> requestDatabase;
    private static ObjectOutputStream requestFileWriter;
    private HttpServletRequest httpRequest;

	public PAPIPoA() {
        PAPIPoA.log = Logger.getLogger(PAPIPoA.class.getName());
	}
	
    public void init(String configFileName)  throws Exception {
    	this.configFileName = configFileName;
        papiConf = new Properties();
        requestDatabase = new HashMap<String,HashMap<String,String>>();
        //read configuration
        FileInputStream configFile = new FileInputStream(this.configFileName);
        papiConf.loadFromXML(configFile);
        
        DOMConfigurator.configure((String)papiConf.get("PoAconf.log4j"));
        
        PAPIPoA.serviceID = (String)papiConf.get("PoAconf.serviceID");
        
        PAPIPoA.location = (String)papiConf.get("PoAconf.location");
        PAPIPoA.log.info("Location: " + PAPIPoA.location);
        
        ciphers = new Ciphers(papiConf);
        
        cookieRegistry = new HashMap<String,Vector<String>>();
        
        //init cookie registry
        String PoApath = (String)(String)papiConf.get("PoAconf.path");
        String cookieFileName = (String)(String)papiConf.get("PoAconf.cookieFile");
        BufferedReader cookieFile = new BufferedReader(new FileReader(PoApath + cookieFileName));
        String line = cookieFile.readLine();
        while (line != null) {
        	String[] data = line.split("\t");
        	Vector<String> values = null;
        	if (cookieRegistry.containsKey(data[0])) {
        		values = (Vector<String>)cookieRegistry.get(data[0]);
        	} else {
        		values = new Vector<String>();
        	}
        	values.add(data[1]);
        	values.add(data[2]);
        	values.add(data[3]);
        	cookieRegistry.put(data[0],values);
        	line = cookieFile.readLine();
        }
        cookieFile.close();
        
        String requestFileName = (String)papiConf.get("PoAconf.requestFile");
        
        File f = new File(PoApath + requestFileName);
        if (f.exists()) {
            FileInputStream requestFile = new FileInputStream(f);
            ObjectInputStream requestInputStream = new ObjectInputStream(requestFile);

            PAPIPoA.requestDatabase = (HashMap<String,HashMap<String,String>>)requestInputStream.readObject();//savedRequest;

            requestInputStream.close();
            requestFile.close();
        }
        else {
        	f.createNewFile();
        	PAPIPoA.requestDatabase = new HashMap<String,HashMap<String,String>>();
        }
        
        
        
        
        
        cookieFileWriter = new BufferedWriter(new FileWriter(PoApath + cookieFileName,true));
        
        
        
    }

    public Map<String,String> doFilter(ServletRequest request, ServletResponse response) throws LoginException, IOException {
        int status = INIT;
        Map<String,String> succeed = null;
        String action = "";
        HashMap oldRequest = null;
        httpRequest = (HttpServletRequest)request;
        HttpServletResponse httpResponse = (HttpServletResponse)response;
        try {
			parameters = new HashMap<String,String[]>(httpRequest.getParameterMap());
            
            if (parameters.containsKey("ACTION")) {
                String[] actions = (String[])parameters.get("ACTION");
                action = actions[0];
                if (action.equals("LOGIN")) {
                    PAPIPoA.log.info("LOGIN query from " + request.getRemoteHost());
                    String path = validateAS(httpRequest);
                    if (!path.equals("")) {
                        ciphers.init(path);
                        status = decodeAssertion(httpRequest);
                    }
                    
                } else if (action.equals("CHECKED")) {
                    PAPIPoA.log.info("CHECKED query from " + request.getRemoteHost());
                    String path = (String)papiConf.get("ParentGPoA.pubkeyName");
                    if (path != null) {
                        status = decodeData(httpRequest, "ParentGPoA");
                        if (status != ACCESS_FATAL) {
                            PAPIPoA.log.info("Recovering query from DB...");
	                        String key = ((String[])parameters.get("KEY"))[0];
	                        if (requestDatabase.containsKey(key)) {
	                            oldRequest = (HashMap)requestDatabase.get(key);
	                            requestDatabase.remove(key);
	                            PAPIPoA.log.info("Saved query successfully recovered and deleted form DB");
	                        }
	                        else {
	                            PAPIPoA.log.fatal("Unable to decode data from CHECKED query");
	                            status = ACCESS_FATAL;
	                        }
                        }
                    }
                    else {
                        PAPIPoA.log.fatal("Invalid public key available for this query");
                        status = ACCESS_FATAL; //aadido a mano, verificar comportamiento
                    }
                    
                } else if (action.equals("TEST")) {
                    PAPIPoA.log.info("TEST query from " + request.getRemoteHost());
                    String path = validateAS(httpRequest);
                    if (!(path.equals("")))
                        status = LOGIN_ACCEPTED;
                    else
                        status = LOGIN_REJECTED;
                    
                } else if (action.equals("LOGOUT")) {
                    status = LOGOUT_ACCEPTED;
                }
                
            } else { //ACCESS
                PAPIPoA.log.info("ACCESS query from " + request.getRemoteHost());
                Cookie[] cookies = httpRequest.getCookies();
                if ( (cookies != null) && (cookies.length > 0) ) {
                    status = validateCookies(cookies);
                } else {
                    status = ACCESS_UNAUTHORIZED;
                }
            }
        
            PAPIPoA.log.info("PoA status: " + String.valueOf(status));
            
            if (status == LOGIN_ACCEPTED) {
                attachCookies(httpRequest, httpResponse, false);
                attachFile(httpResponse, true);
            } else if (status == LOGIN_REJECTED) {
                attachFile(httpResponse, false);
            } else if (status == LOGOUT_ACCEPTED) {
                attachCookies(httpRequest, httpResponse, true);
                attachFile(httpResponse, true);
            } else if ((status == ACCESS_ACCEPTED) && (action.equals("CHECKED")) && oldRequest.get("METHOD").equals("POST")) {
                attachCookies(httpRequest, httpResponse, false);
                attachPOSTParams(httpResponse, oldRequest);
                succeed = createUserInfo();
            } else if ((status == ACCESS_ACCEPTED) && (action.equals("CHECKED")) ) {
                attachCookies(httpRequest, httpResponse, false);
                succeed = createUserInfo();
                PAPIPoA.log.info("Redirecting user to original location...");
                httpResponse.sendRedirect((String)oldRequest.get("REQUEST_URI"));
            } else if (status == ACCESS_ACCEPTED) {
                attachCookies(httpRequest, httpResponse, false);
                succeed = createUserInfo();
            } else if ( (status == ACCESS_UNAUTHORIZED) && papiConf.containsKey("ParentGPoA.URL") && !(papiConf.getProperty("ParentGPoA.URL").equals(""))){
                String key = String.valueOf(System.currentTimeMillis());
                //save request in internal database
                HashMap<String,String> clonedRequest = cloneRequest(httpRequest);
                requestDatabase.put(key,clonedRequest);
                PAPIPoA.log.info("Request successfully saved in DB");
                String gpoaURLToken = (String)papiConf.get("ParentGPoA.URL");
                PAPIPoA.log.info("Redirecting to parent GPoA");
                httpResponse.sendRedirect(gpoaURLToken + "?ACTION=CHECK" + "&DATA=" + key + "&URL=" + httpRequest.getRequestURL().toString()); //+ httpRequest.getQueryString());
            } else {
                httpResponse.sendError(HttpServletResponse.SC_FORBIDDEN, (String)papiConf.get("PoAconf.rejectMessage"));
            }
        } catch (Exception e) {
        	e.printStackTrace();
            httpResponse.sendError(HttpServletResponse.SC_FORBIDDEN, "Exception due to corrupt data");
            PAPIPoA.log.fatal("Exception while handling request. More information about this error may be available in the server error log.");
        }
        
        requestFileWriter = new ObjectOutputStream(new FileOutputStream ((String)papiConf.get("PoAconf.path") + (String)papiConf.get("PoAconf.requestFile")));
        requestFileWriter.writeObject(PAPIPoA.requestDatabase);
        requestFileWriter.close();
        return succeed;
    }
    
    /** Decides wether the PoA recognizes the AS or not
     * @param request Incoming request
     * @return Name of the AS
     * @throws ServletException If the AS is not recognized
     */
    private String validateAS(HttpServletRequest request) {
        String ASname = request.getParameter("AS");
        if ((String)papiConf.get(ASname + ".pubkeyName") != null) {
            return ASname;
        } else
            return "";
    }
    
    /** Decodes and splits the information contained in the 5-field DATA parameter of the request
     * @param request Incoming request
     * @throws ServletException Propagates any exception ocurred while decoding
     */
    private int decodeAssertion(HttpServletRequest request) throws ServletException {
        
        String userAssertion = (String)ciphers.RSA_Encode(false, (String)request.getParameter("DATA"));

        PAPIPoA.log.info("Received assertion: " + userAssertion);

        
        String[] splittedAssertion = userAssertion.split(":");
        boolean validUser = false;
        boolean validServiceID = false;
        boolean onTime = false;
        boolean validLocation = false;
        
        if (request.getParameter("ACTION").equals("LOGIN")) {
            int length = splittedAssertion.length;
            String serviceID = (String)splittedAssertion[length-1];
            String location = (String)splittedAssertion[length-2];
            String currentTimeStr = (String)splittedAssertion[length-3];
            String userData = "";
            for (int i=0; i < length-4; i++) {
                userData = userData.concat((String)splittedAssertion[i]);
            }

            String[] userDataArray = new String[1];
            userDataArray[0] = userData;
            parameters.put("userData", userDataArray);
            long currentTime = Long.parseLong(currentTimeStr)*1000;
            long url_timeout = Long.parseLong((String)(String)papiConf.get("PoAconf.URL_Timeout"));
            long milis = System.currentTimeMillis();
            onTime = milis < (currentTime + url_timeout);
            
            validUser = validateUser(userData,false);
            validLocation = location.equals(PAPIPoA.location);
            
            validServiceID = PAPIPoA.serviceID.equals(serviceID);//splittedAssertion[4]
            
        }
        if (onTime && validServiceID && validLocation && validUser)
            return LOGIN_ACCEPTED;
        else
            return LOGIN_REJECTED;
        
    }
    
    /** Decodes and splits the information contained in the 4-field DATA parameter of the request
     * @param request Incoming request
     * @throws ServletException Propagates any exception ocurred while decoding
     */
    public int decodeData(HttpServletRequest request, String producer) throws ServletException {
        
        boolean validUser = false;
        boolean validServiceID = false;
        boolean onTime = false;

        ciphers.init(producer);
        
		parameters = new HashMap<String,String[]>((HashMap<String,String[]>)request.getParameterMap());
        
        String param =  (String)request.getParameter("DATA");
/*        try {
			param = URLDecoder.decode(param,"UTF-8");
		} catch (UnsupportedEncodingException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
*/        
        String userAssertion = (String)ciphers.RSA_Encode(false,param);
        PAPIPoA.log.info("Received assertion from GPoA: " + userAssertion); //PAPIPoA.log.info
        String[] splittedAssertion = userAssertion.split(":");
		if (((String)splittedAssertion[0]).equals("ERROR")) {
		    return ACCESS_FATAL;
		}
	
        if (request.getParameter("ACTION").equals("CHECKED")) {
            
            int length = splittedAssertion.length;
            String key = (String)splittedAssertion[length-1];
            String[] keyArray = new String[1];
            keyArray[0] = key;
            parameters.put("KEY",keyArray);
            String currentTimeStr = (String)splittedAssertion[length-2];
            long currentTime = Long.parseLong(currentTimeStr)*1000;
            long url_timeout = Long.parseLong((String)papiConf.get("PoAconf.URL_Timeout"));
            long milis = System.currentTimeMillis();
            onTime = milis < (currentTime + url_timeout);
            
            String userData = "";
            for (int i=0; i < length-3; i++) {
                userData = userData + (String)splittedAssertion[i] + ":";
            }

            userData = userData.substring(0, userData.length()-1);
            String[] userDataArray = new String[1];
            userDataArray[0] = userData;
            parameters.put("userData", userDataArray);
            
            validUser = validateUser(userData,false);
        }
        PAPIPoA.log.info("Ontime: " + String.valueOf(onTime));
        PAPIPoA.log.info("validUser: " + String.valueOf(validUser));

        if (onTime && validUser)
            return ACCESS_ACCEPTED;
        else {
            return ACCESS_FATAL;
        }
        
    }
    
    
    
    
    /** Attaches an accept/reject file defined in the profile to the response
     * @param response Response to attach the file to
     * @param accepted If the file to attach is the accept or reject file
     * @throws IOException While working with the file
     */
    private void attachFile(HttpServletResponse response, boolean accepted) throws IOException {
        String filename = "";
        if (accepted)
            filename = (String)(String)papiConf.get("PoAconf.acceptURL");
        else
            filename = (String)(String)papiConf.get("PoAconf.rejectURL");
        try {
            File file = new File(filename);
            FileReader reader = new FileReader(file);
            Long fileSize = new Long(file.length());
            int fileLength = fileSize.intValue();
            response.setContentLength( fileLength);
            char[] output = new char[fileLength];
            reader.read(output);
            OutputStream stream = response.getOutputStream();
            String strOutput = new String(output);
            stream.write(strOutput.getBytes());
        } catch (Exception e) {
            throw new IOException(e.getMessage());
        }
        
    }
    
    /** Attaches a pair of valid cookies (Hcook and Lcook) to the response
     * @param request Request with values to create the cookies
     * @param response HttpResponse
     * @param outdated If we have to create outdated cookies, as in a LOGOUT handling
     * @throws ServletException If there has been an exception while encoding the cookies
     */
    private void attachCookies(HttpServletRequest request, HttpServletResponse response, boolean outdated) throws ServletException {
        Cookie hcook = new Cookie("Hcook", ciphers.AES_Encode(true, true, generateHcook(request,outdated)));
        String maxAge = (String)(String)papiConf.get("PoAconf.cookieMaxAge");
        if (outdated) {
            maxAge = "0";
        }
        hcook.setMaxAge(Integer.parseInt(maxAge));
        hcook.setPath(location);
        response.addCookie(hcook);
        Cookie lcook = new Cookie("Lcook", ciphers.AES_Encode(true, false, generateLcook(request,outdated)));
        lcook.setPath(location);
        response.addCookie(lcook);
    }
    
    /** Generates the information for the Hcook
     * @param request Incoming request
     * @param outdated If we have to create outdated cookies, as in a LOGOUT handling
     * @return Hcook data
     * @throws ServletException if there has been an exception while writing in the cookie registry
     */
    private String generateHcook(HttpServletRequest request, boolean outdated) throws ServletException {
        String rawCook = "";
        rawCook += new Boolean( (String)papiConf.get("ParentGPoA.URL") != null).toString();
        rawCook += ":";
        String userData = ((String[])parameters.get("userData"))[0];
        rawCook += userData;
        rawCook += ":";
        String maxAge = (String)(String)papiConf.get("PoAconf.cookieMaxAge");
        long expiryTime = 0;
        if (!outdated) {
            expiryTime = (System.currentTimeMillis())/1000  + Integer.parseInt(maxAge);
        }
        rawCook += String.valueOf(expiryTime);
        rawCook += ":";
        String randomBlock = String.valueOf(System.currentTimeMillis());
        rawCook += randomBlock;
        rawCook += ":";
        Vector<String> cookieValues = new Vector<String>();
        cookieValues.add(location);
        cookieValues.add(userData);
        cookieValues.add(serviceID);
        cookieRegistry.put(randomBlock, cookieValues);
        try {
            cookieFileWriter.append(randomBlock + "\t" + location + "\t" + userData + "\t" + serviceID + "\n");
            cookieFileWriter.flush();
        } catch (IOException e) {throw new ServletException(e.getMessage());}
        
        rawCook += location;
        rawCook += ":";
        String serviceID = PAPIPoA.serviceID;
        rawCook += serviceID;
        
        return rawCook;
    }
    
    /** Generates the information for the Lcook
     * @param request Incoming request
     * @param outdated If we have to create outdated cookies, as in a LOGOUT handling
     * @return Lcook data
     * @throws ServletException if there has been an exception while writing in the cookie registry
     */
    private String generateLcook(HttpServletRequest request, boolean outdated) {
        String rawCook = "";
        long timeStamp = System.currentTimeMillis()/1000;
        if (outdated) {
            timeStamp = 0;
        }
        rawCook += String.valueOf(timeStamp);
        rawCook += ":";
        rawCook += PAPIPoA.location;
        rawCook += ":";
        String serviceID = PAPIPoA.serviceID;
        rawCook += serviceID;
        rawCook += ":";
        String userData = ((String[])parameters.get("userData"))[0];
        rawCook += userData;
        return rawCook;
    }
    
    /** Decodes and validates the information contained in the cookies
     * @param cookies Incoming request's cookie array
     * @throws ServletException Propagates any exception ocurred while decoding
     */
    private int validateCookies(Cookie[] cookies) throws ServletException {
        int HcookIndex = 0;
        int LcookIndex = 0;
        String data = "";
        int stat = ACCESS_UNAUTHORIZED;
        
        int cookiesLength = cookies.length;
        while (  (LcookIndex < cookiesLength) && (!(cookies[LcookIndex].getName().equals("Lcook")))){
            LcookIndex++;
        }
        
        if (LcookIndex < cookies.length) {
            data = ciphers.AES_Encode(false, false, cookies[LcookIndex].getValue());
            stat = validateCookie(false, data);
            if (stat == ACCESS_ACCEPTED) {
                return stat;
            }
        }
        
        while ((HcookIndex < cookiesLength) && !(cookies[HcookIndex].getName().equals("Hcook"))){
            HcookIndex++;
        }
        if (HcookIndex < cookies.length) {
            data = ciphers.AES_Encode(false, true, cookies[HcookIndex].getValue());
            stat = validateCookie(true, data);
        }
        return stat;
    }
    
    
    /** Validates the information contained in a cookie
     * @param isHcook true if it's a Hcook, false if it's a Lcook
     * @param data Cookie data
     * @return Status code derived from validation process
     */
    private int validateCookie(boolean isHcook, String data) {
        
        String[] splittedValues = data.split(":");
        
        if (isHcook) {
        	
        	httpRequest.getSession().setAttribute(this.getClass().getName() + ".PAPIHcookValue",data);
        	
            int length = splittedValues.length;
            String isTemporary = (String)splittedValues[0];
            String serviceID = (String)splittedValues[length-1];
            String location = (String)splittedValues[length-2];
            String randomBlock = (String)splittedValues[length-3];
            String expiryTime = (String)splittedValues[length-4];
            String userData = "";
            for (int i=1; i < length-4; i++) {
                userData = userData.concat((String)splittedValues[i]);
            }

            String[] userDataArray = new String[1];
            userDataArray[0] = userData;
            parameters.put("userData", userDataArray);
            
            long expTime = Long.parseLong(expiryTime)*1000;
            long currentTime = System.currentTimeMillis() ;
            boolean onTime = currentTime >=  expTime;
            if (currentTime >=  expTime) {
                return ACCESS_UNAUTHORIZED;
            }
            
            Vector blocks = (Vector)cookieRegistry.get(randomBlock);

            boolean validRandomBlock = !(blocks.equals(null)) && blocks.contains(location) && blocks.contains(userData) && blocks.contains(serviceID);
            boolean validLocation = location.equals(PAPIPoA.location);
            boolean validServiceID = PAPIPoA.serviceID.equals(serviceID);
            boolean validUser = validateUser(userData, true);
            
            if (validRandomBlock && validLocation && validServiceID && validUser)
                return ACCESS_ACCEPTED;
            else
                return ACCESS_FATAL;
            
        } else { //lcook
            
            int length = splittedValues.length;
            String timeStampStr = (String)splittedValues[0];
            String location = (String)splittedValues[1];
            String serviceID = (String)splittedValues[2];
            
            String userData = "";
            for (int i=3; i < length; i++) {
                userData = userData.concat((String)splittedValues[i]);
            }
            
            String[] userDataArray = new String[1];
            userDataArray[0] = userData;
            parameters.put("userData", userDataArray);
            
            long currentTime = System.currentTimeMillis();
            long timeStamp = Long.parseLong(timeStampStr)*1000;
            String timeOut = (String)papiConf.get("PoAconf.Lcook_Timeout");
            boolean onTime = currentTime < (timeStamp + Long.parseLong(timeOut));
            if (!onTime) {
                return ACCESS_UNAUTHORIZED;
            }
            
            boolean validLocation = location.equals(PAPIPoA.location);
            boolean validServiceID = PAPIPoA.serviceID.equals(serviceID);
            boolean validUser = validateUser(userData,true);

            if (validLocation && validServiceID && validUser)
                return ACCESS_ACCEPTED;
            else 
                return ACCESS_FATAL;
        }
        
    }
    
    /** Stores the relevant parameters of a request in a Map.
     *  @param request Request to clone
     *  @return Map associating a request parameter and its value
     */
    private HashMap<String,String> cloneRequest(HttpServletRequest request) {
        HashMap<String,String> params = new HashMap<String,String>();
        
        params.put("METHOD", request.getMethod());
        params.put("PATH_INFO", request.getPathInfo());
        params.put("REQUEST_URI", request.getRequestURI());
        params.put("QUERY_STRING", request.getQueryString());
        if (request.getMethod().equals("POST")) {
            Enumeration e = request.getParameterNames();
            while (e.hasMoreElements()) {
                String key = (String)e.nextElement();
                String value = request.getParameter(key);
                params.put("POST." + key, value);
            }
        }
        return params;
    }
    
    /** Attaches the POST parameters saved in a previous request to the actual one
     *  @param response HTTP response
     *  @param oldRequest Map with old request's values
     *  @throws IOException If error while writing data on the response
     *  @throws ServletException If incorrect values on configuration file
     */
    private void attachPOSTParams(HttpServletResponse response, HashMap oldRequest) throws IOException, ServletException {
        response.setContentType("text/html");
        OutputStream output = response.getOutputStream();
        String preamble = "";
        if (((String)papiConf.get("PoAconf.POST_Method")).equals("javascript")) {
            preamble = "<html>" +
                "<head><title>PAPI Internal Location Service</title></head>" +
                "<body onload=proxy.submit()><h3>Please wait while you are being redirected...</h3>" +
                "<form name=proxy method=\"post\" action=\"" + (String)oldRequest.get("REQUEST_URI") + "\">";
        } else if (((String)papiConf.get("PoAconf.POST_Method")).equals("manual")) {
            preamble = "<html>" +
                "<head><title>PAPI Internal Location Service</title></head>" +
                "<body><h3>Press the \"Submit\" button to proceed with the request...</h3>" +
                "<form name=proxy method=\"post\" action=\"" + (String)oldRequest.get("REQUEST_URI") + "\">";
        } else
            throw new ServletException ("Invalid value for POST_Method option in the config file.");
        
        output.write(preamble.getBytes());
        
        Iterator keys = oldRequest.keySet().iterator();
        while (keys.hasNext()) {
            String key = (String)keys.next();
            if (key.startsWith("POST.")) {
                String value = (String)oldRequest.get(key);
                key = key.substring(5,key.length());
                String input = "<input name=" + key + " type=hidden value=" + value + " />";
                output.write(input.getBytes());
            }
        }

        String end = "";
        if (((String)papiConf.get("PoAconf.POST_Method")).equals("javascript")) {
            end = "</form></body>" + "</html>";
        } else if (((String)papiConf.get("PoAconf.POST_Method")).equals("manual")) {
            end = "<input type=\"submit\" value=\"Submit\"/>" + "</form></body>" + "</html>";
        }
        output.write(end.getBytes());
        
        output.flush();
        output.close();
        
    }
    
    /** Validates a user/group according to the PAPI_Filter or Cookie_Reject values
     *  @param userData Data provided by the assertion
     *  @param isCookie Defines which filter to apply (cookie or general filter)
     *  @return True if the user/group has access rights
     */
    private boolean validateUser (String userData, boolean isCookie) {
	        boolean validUser = false;
	        String filterToken = "";
	        if (isCookie)
	            filterToken = (String)papiConf.get("PoAconf.Cookie_Reject");
	        else
	            filterToken = (String)papiConf.get("PoAconf.PAPI_Filter");
	            
	        String[] assertion = userData.split("@");
	        String userAS = (String)assertion[assertion.length-1];
//	    	httpRequest.getSession().setAttribute(this.getClass().getName() + ".PAPIAuthServer",userAS);
	        parameters.put(this.getClass().getName() + ".PAPIAuthServer",new String[]{userAS});
	        
	        String userValues = "";
	        for (int i=0; i<assertion.length-1; i++) {
	            userValues = userValues + assertion[i] + "@";
	        }
	        
	        userValues = userValues.substring(0, userValues.length()-1);

	        String attSeparator = (String)papiConf.get("PoAconf.attSeparator");
	        String[] attributes = userValues.split(attSeparator);
	        
	        String attValSeparator = (String)papiConf.get("PoAconf.attValSeparator");
	        

	        
	        for (int i=0; i<attributes.length; i++) {
	            String[] splittedAttr = attributes[i].split(attValSeparator);
	            String attr = this.getClass().getName() + "." + splittedAttr[0].trim(); //this.getClass().getName() + "." + 
	            String val = "";
	            if (splittedAttr.length > 1)
	            	val = splittedAttr[1].trim();

	            String[] values = new String[1];
	            values[0] = val;
	            parameters.put(attr, values);
//	            httpRequest.getSession().setAttribute(attr, val);
	        }

	        int separator = filterToken.indexOf(">");
	        String filteredRole = filterToken.substring(0, separator - 1);
	        filteredRole = filteredRole.trim();
	        String filter = filterToken.substring(separator + 1, filterToken.length());
	        filter = filter.trim();
	        String groupToken = "";
	        String uidToken = "";

	        if (groupToken.matches(filteredRole) || uidToken.matches(filteredRole) || filteredRole.equals("any")) {
	            if (filter.equals("accept"))
	                validUser = true;
	            else
	                validUser = false;
	        }
	        
	        return validUser;
	        
	    }   
    
    private Map<String,String> parseData (String userData) {
    	Map<String,String> parsedData = new HashMap<String,String>();

        String[] assertion = userData.split("@");
        String userAS = (String)assertion[assertion.length-1];
    	parsedData.put(this.getClass().getName() + ".PAPIAuthServer",userAS);

        String userValues = "";
        for (int i=0; i<assertion.length-1; i++) {
            userValues = userValues.concat(assertion[i]);
        }
        
        String attSeparator = (String)papiConf.get("PoAconf.attSeparator");
        String[] attributes = userValues.split(attSeparator);
        
        String attValSeparator = (String)papiConf.get("PoAconf.attValSeparator");
        

        
        for (int i=0; i<attributes.length; i++) {
            String[] splittedAttr = attributes[i].split(attValSeparator);
            String attr = this.getClass().getName() + "." + splittedAttr[0].trim();
            String val = splittedAttr[1].trim();
        	parsedData.put(attr, val);
        }

    	
    	return parsedData;
    }

    private Map<String,String> createUserInfo() {
    	Map<String,String> info = new HashMap<String,String>();
    	String data = ((String[])parameters.get("userData"))[0];
    	info = parseData(data);
    	return info;
    }
    
    public String getOption(String key) throws LoginException {
    	if (!PAPIPoA.papiConf.containsKey(key))
    		throw new LoginException ("Unknown option: " + key);
    		
    	else
    		return PAPIPoA.papiConf.getProperty(key);
    }
}
